#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include "dictre.h"

const static char indexext[] = ".index";
const static char dictext[] = ".dict";

int get_filesize(const char *fname)
{
    struct stat st;
    int res;
    res = stat(fname, &st);
    if (res == -1)
    {
        fprintf(stderr, "Problem stating the file %s\n", fname);
        perror("Error:");
        exit(-1);
    }

    return st.st_size;
}

void init_dictionary(struct Dict *d, const char *base)
{
    char *filename;

    filename = (char *) malloc(strlen(base) + 10);

    /* Prepare .index filename and open it*/
    strcpy(filename, base);
    strcat(filename, indexext);

    d->indexsize = get_filesize(filename);
    d->indexfd = open(filename, O_RDONLY);
    if (d->indexfd == -1)
    {
        fprintf(stderr, "Problem opening the file %s\n", filename);
        perror("Error:");
        exit(-1);
    }
    d->index = (unsigned char *) mmap(0, d->indexsize, PROT_READ, MAP_SHARED,
            d->indexfd, 0);

    /* Prepare .dict filename and open it*/
    strcpy(filename, base);
    strcat(filename, dictext);
    d->defs = fopen(filename, "r");
    if (d->defs == 0)
    {
        fprintf(stderr, "Problem opening the file %s\n", filename);
        perror("Error:");
        exit(-1);
    }

    free(filename);
}

void end_dictionary(struct Dict *d)
{
    munmap(d->index, d->indexsize);
    close(d->indexfd);
    fclose(d->defs);
}

static void fill_def(struct Dict *d, int offset, int length, char * def)
{
    fseek(d->defs, offset, SEEK_SET);
    fread(def, 1, length, d->defs);
}

static int pointer_at_end(struct Dict *d, unsigned char *ptr)
{
    if (ptr >= (d->index + d->indexsize))
        return 1;
    return 0;
}

static char * skip_until_newline(struct Dict *d, char *from)
{
    if (pointer_at_end(d, from))
        return 0;
    while(*from != '\n' && *from != 0)
    {
        ++from;
        if(pointer_at_end(d, from))
            return 0;
    }
    return from;
}

static int compare(const unsigned char *word, const unsigned char *test)
{
    int i;

    /*printf("Comparing %s to %.20s\n", word, test);*/
    for(i=0; word[i] != 0 && test[i] != 0; ++i)
    {
        if (word[i] != test[i])
        {
            break;
        }
    }
    if (word[i] == 0 && test[i] == '\t')
        return 0;
    else if (word[i] == 0)
        return -1;
    else if (test[i] == '\t')
        return 1;
    else if (word[i] > test[i])
        return 1;
    else if (word[i] < test[i])
        return -1;

    /* It should never reach this. */
    return -1;
}

static char * bin_search(struct Dict *d, const char *word)
{
    int step, pivot;

    pivot = d->indexsize / 2;
    step = d->indexsize / 2;

    do
    {
        char *test;
        int comparision;
        test = d->index + pivot;
        test = skip_until_newline(d, test);
        if (test == 0)
            return 0;
        test += 1; /* skip exactly the new line */

        comparision = compare(word, test);
        if (comparision == 0)
        {
            return test + strlen(word) + 1; /* skip word and \n */
        } else if (comparision < 0)
        {
            step = step / 2;
            pivot = pivot - step;
        } else if (comparision > 0)
        {
            step = step / 2;
            pivot = pivot + step;
        }
    } while(step > 0);
    return 0;
}

static int my_get_int(char **pos)
{
    int i;
    char *start;
    int val;

    start = *pos;
    for(i=0; start[i] != '\t' && start[i] != '\n'; ++i)
        ;
    val = str2int_len(start, i);
    *pos += i + 1;
    return val;
}

void find_def(struct Dict *d, const char *word, char * def)
{
    int offset, len;
    char *pos;

    pos = bin_search(d, word); /* pos points to the offset already. */
    if (pos == 0)
    {
        def[0] = 0;
        /*fprintf(stderr, "Cannot find %s\n", word);*/
        return;
    }
    offset = my_get_int(&pos); /* increments pos */
    len = my_get_int(&pos); /* increments pos */
    fill_def(d, offset, len, def);
}
